/* TitraxConduit.cpp
   function implementation for a DLL that can be used on windows systems to synchronize the
   the titrax palm program with the desktop system via the Palm CDK API's.

   Copyright Jim Weller 2004

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA. 
*/

#include "TitraxConduit.h"


/*
*****************************************************************************
*
*     Function:      DllMain()
*
*     Description:   main entry point to the TitraxConduit component
*
*     Parameters(3):
*                    hInstance  - instance handle of the DLL
*                    dwReason   - why the entry point was called
*                    lpReserved - reserved
*
*     Returns:       1 if okay
*
*****************************************************************************
*/
extern "C" int APIENTRY DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
{

    if (dwReason == DLL_PROCESS_ATTACH)
    {
        TRACE0("EXPORTED FUNCTIONS Initializing!\n");
        _tzset(); // init timezone
        if ( !iDllAttachCount ) {
            hAppInstance = hInstance;
            // PalmLoadLanguage 
            hLangInstance = hInstance;
        }
        ++iDllAttachCount;
    }
    else if (dwReason == DLL_PROCESS_DETACH)
    {
        TRACE0("EXPORTED FUNCTIONS Terminating!\n");

        --iDllAttachCount;
        if (!iDllAttachCount ) {
            // PalmFreeLanguage
        }
    }
    return 1;   // ok
}



/*
*****************************************************************************
*     EXTERN
*
*     Function:        OpenConduit()
*
*     Description:  Extern "C" entry point into this conduit which starts the
*                   process of synchronizing
*
*     Parameters(2):   Pointer to a callback function used to report progress.
*                      Sync settings/properties from HotSync
*                    
*     Returns:  !0 - Error
*
*****************************************************************************
*/
ExportFunc long OpenConduit(PROGRESSFN progress,CSyncProperties &sync)
{
  long err =0;
  CONDHANDLE myConduitHandle =(CONDHANDLE)0;
  TCHAR conduitName[MAX_PATH]; 

  titraxReadRegistry();
  LogAddFormattedEntry(slText,false, "summ?%ld,hist?%ld,file?%ld,time?%ld",
                                      titraxDumpSummary,
                                      titraxDumpHistory,
                                      titraxFileTypeOption,
                                      titraxTimeTypeOption
                                        );


  GetConduitName(conduitName,sizeof(conduitName));
  
  if (!progress){
    LogAddFormattedEntry(slText,false, "%s - an improper progress function pointer was passed to OpenConduit",conduitName);
    return -1;
  }

  if (sync.m_SyncType == eDoNothing) { 
    LogAddFormattedEntry(slText,false, "%s - sync configured to Do Nothing",conduitName);
    return 0;
  }

  err = SyncRegisterConduit(myConduitHandle);

  if( err == 0 ){
    LogAddEntry("",slSyncStarted, false);

    switch(sync.m_SyncType){

        case eFast:
        case eSlow:
            LogAddFormattedEntry(slText,false, "%s - palm<->pc",conduitName);
            loadTitraxRecords();
            if( titraxDumpHistory )
                titraxDumpHistoryToFile(sync.m_PathName);
            if( titraxDumpSummary )
                titraxDumpSummaryToFile(sync.m_PathName);
            break;

        case eHHtoPC:
            LogAddFormattedEntry(slText,false, "%s - palm->pc (not implemented)",conduitName);
            break;
  
        case ePCtoHH:
            LogAddFormattedEntry(slText,false, "%s - palm<-pc (not implemented)",conduitName);
            break;

        default:
            LogAddFormattedEntry(slText,false, "%s - syncType unknown",conduitName);
            break;
    }

    LogAddEntry(conduitName,err ? slSyncAborted : slSyncFinished, false);
    SyncUnRegisterConduit(myConduitHandle);
  }

  return err;
}



/*
*****************************************************************************
*       EXTERN
*
*       Function:     GetConduitName()
*
*       Description:  Extern "C" entry point into this conduit which returns
*                     the name to be used when displaying messages regarding
*                     this conduit.
*
*       Parameters(2):   pszName - buffer in which to place the name
*                        nLen - maximum number of bytes of buffer     
*                                      
*       Returns:      !0 - error
*
*****************************************************************************
*/
ExportFunc long GetConduitName(TCHAR* pszName,WORD nLen)
{
    long retval = -1;

    if (LoadString((HINSTANCE)hLangInstance, IDS_CONDUIT_NAME, pszName, nLen))
        retval = 0;

    return retval;
}


/*
*****************************************************************************
*       EXTERN
*
*       Function:     GetConduitVersion()
*
*       Description:  Extern "C" entry point into this conduit which returns
*                     the conduits version
*
*       Parameters:   none
*
*       Returns:      DWORD indicating major and minor version number
*                        HIWORD - reserved
*                        HIBYTE(LOWORD) - major number
*                        LOBYTE(LOWORD) - minor number
*
*****************************************************************************
*/
ExportFunc DWORD GetConduitVersion()
{
    return TITRAX_CONDUIT_VERSION;
}

/*
*****************************************************************************
*       EXTERN
*
*       Function:     ConfigureConduit
*
*       Description:  Extern "C" entry point into this conduit which 
*                     displays the UI necessary to configure this 
*                     conduit.
*
*       Parameters(1):  CSyncPreference - sync prefs from hotsync
*
*       Returns:      0 - success, !0 - failure
*
*****************************************************************************
*/
ExportFunc long ConfigureConduit(CSyncPreference& pref)
{

    long nRtn = -1;
    CfgConduitInfoType cfg;
    cfg.dwVersion = CFGCONDUITINFO_VERSION_1;
    cfg.dwSize  = sizeof(CfgConduitInfoType);
    cfg.dwCreatorId = 0;
    cfg.dwUserId = 0;
    memset(cfg.szUser , 0, sizeof(cfg.szUser));
    memset(cfg.m_PathName, 0, sizeof(cfg.m_PathName)); 
    cfg.syncPermanent = pref.m_SyncType;
    cfg.syncTemporary = pref.m_SyncType;
    cfg.syncNew = pref.m_SyncType;
    cfg.syncPref = eTemporaryPreference; 

    int iResult;
    iResult = DialogBoxParam((HINSTANCE)hLangInstance, 
              MAKEINTRESOURCE(IDD_TITRAXCONDUIT_DIALOG), 
              GetForegroundWindow(), 
              (DLGPROC)HSConfigurationDialogProc,
              (LPARAM)&cfg);
    if (iResult == 0) {
        pref.m_SyncType = cfg.syncNew;
        pref.m_SyncPref = cfg.syncPref;
        nRtn = 0;
    }
    return nRtn;
}

/*
*****************************************************************************
*
*     Method:        GetConduitInfo
*
*     Description:    This function provides a way for a Conduit to provide info
*                  to the caller (name, MFC version, and default sync action)
*
*     Parameters(4):  ConduitInfoEnum infoType - enum specifying what info is being
*                                             requested.
*                  void *pInArgs - This parameter may be null, except for the Conduit
*                                  name enum, this value will be a ConduitRequestInfoType structure.
*
*                  pOut and pOutSize depending on enum infoType
*                     For eConduitName
*                         void *pOut - will be a pointer to a character buffer
*                         DWORD *pdwOutSize - will be a pointer to a DWORD specifying the size of the character buffer.
*
*                     For eMfcVersion
*                         void *pOut - will be a pointer to a DWORD
*                         DWORD *pdwOutSize - will be a pointer to a DWORD specifying the size of pOut.
*
*                     For eDefaultAction
*                         void *pOut - will be a pointer to a eSyncType variable
*                         DWORD *pdwOutSize - will be a pointer to a DWORD specifying the size of pOut.
*
*     Returns:        0        - Success.
*                    !0      - error code.
*
*****************************************************************************
*/
ExportFunc long GetConduitInfo(ConduitInfoEnum infoType, void *pInArgs, void *pOut, DWORD *pdwOutSize)
{
    if (!pOut)
        return CONDERR_INVALID_PTR;
    if (!pdwOutSize)
        return CONDERR_INVALID_OUTSIZE_PTR;

    switch (infoType) {
        case eConduitName:
            if (!pInArgs)
                return CONDERR_INVALID_INARGS_PTR;
            ConduitRequestInfoType *pInfo;
            pInfo = (ConduitRequestInfoType *)pInArgs;
            if ((pInfo->dwVersion != CONDUITREQUESTINFO_VERSION_1) ||
                (pInfo->dwSize != SZ_CONDUITREQUESTINFO))
                return CONDERR_INVALID_INARGS_STRUCT;
            
            if (!LoadString((HINSTANCE)hLangInstance, IDS_CONDUIT_NAME, (TCHAR*)pOut, *pdwOutSize))
                return CONDERR_CONDUIT_RESOURCE_FAILURE;
            break;
        case eDefaultAction:
            if (*pdwOutSize != sizeof(eSyncTypes))
                return CONDERR_INVALID_BUFFER_SIZE;
            (*(eSyncTypes*)pOut) = eFast;
            break;
        case eMfcVersion:
            if (*pdwOutSize != sizeof(DWORD))
                return CONDERR_INVALID_BUFFER_SIZE;
            (*(DWORD*)pOut) = MFC_NOT_USED;
            break;
        default:
            return CONDERR_UNSUPPORTED_CONDUITINFO_ENUM;
    }
    return 0;
}


/*
*****************************************************************************
*    EXTERN
*
*     Method:        CfgConduit
*
*     Description:    This is the second instance of ConfigureConduit. This function
*                  is provided with more info from the caller.
*                  This version of the function only supports the eConfig1 enum.
*
*     Parameters(3):  ConduitCfgEnum cfgType - enum specifying the version of the CfgConduit Call.
*                  void *pArgs - In the eConfig1 call, this is a pointer to aCfgConduitInfoType 
*                                structure. This structure provides extra info to the conduit.
*                  DWORD *pdwArgsSize - This is a pointer to a DWORD specifying the size of pArgs.
*
*     Returns:       !0      - error code.
*
*****************************************************************************
*/
ExportFunc long CfgConduit( ConduitCfgEnum cfgType, void *pArgs, DWORD *pdwArgsSize)
{
    long nRtn = -1;
    TCHAR szName[256]; // yuck magic number
    DWORD dwNamesize;
    ConduitRequestInfoType infoStruct;
    CfgConduitInfoType *pCfgInfo;

    dwNamesize = sizeof(szName);

    if (!pArgs)
        return CONDERR_INVALID_INARGS_PTR;
    if (!pdwArgsSize)
        return CONDERR_INVALID_ARGSSIZE_PTR;
    if (*pdwArgsSize != SZ_CFGCONDUITINFO) 
        return CONDERR_INVALID_ARGSSIZE;

    if (cfgType != eConfig1)
            return CONDERR_UNSUPPORTED_CFGCONDUIT_ENUM;

    pCfgInfo = (CfgConduitInfoType *)pArgs;
    if (pCfgInfo->dwVersion != CFGCONDUITINFO_VERSION_1)
        return CONDERR_UNSUPPORTED_STRUCT_VERSION;



    infoStruct.dwVersion = CONDUITREQUESTINFO_VERSION_1;
    infoStruct.dwSize = SZ_CONDUITREQUESTINFO;
    infoStruct.dwCreatorId = pCfgInfo->dwCreatorId;
    infoStruct.dwUserId = pCfgInfo->dwUserId;
    StringCchCopy(infoStruct.szUser,sizeof(infoStruct.szUser),pCfgInfo->szUser);
    nRtn = GetConduitInfo(eConduitName, (void *)&infoStruct, (void *)szName, &dwNamesize);
    if (nRtn)
        return nRtn;

    int iResult;
    iResult = DialogBoxParam((HINSTANCE)hLangInstance, 
              MAKEINTRESOURCE(IDD_TITRAXCONDUIT_DIALOG), 
              GetForegroundWindow(), 
              (DLGPROC)HSConfigurationDialogProc,
              (LPARAM)pCfgInfo);

    return 0;
}




/*
*****************************************************************************
*
*     Method:        HSConfigurationDialogProc
*
*     Description:    This is a callback to be called given to windows from the 
*                     conduit configuration functions. It displays the titrax
*                     configuration dialog. It is also the callback
*                     for user events in the dialog. It reads and writes
*                     user settings to the registry.
*
*     Parameters(3):    HWND hWnd - app handle
*                       UINT Message - type of action/event
*                       LPARAM lParam - info about the message
*
*     Returns:       !0 - error
*
*****************************************************************************
*/
long CALLBACK HSConfigurationDialogProc(HWND hWnd, UINT Message, WPARAM wParam, LPARAM lParam)
{
    static CfgConduitInfoType *pCfgInfo;
    TCHAR szPath[256];
    boolean needRegWrite=FALSE;


    // why/how were we called?
    switch (Message) 
    {
        case WM_INITDIALOG: // Somebody pressed "Setup..." in the Hotsync's "Custom..." area. Build & show dialogue.
            titraxReadRegistry();
            switch( titraxFileTypeOption ){
                case csv_file:
                    CheckRadioButton( hWnd, IDC_RADIO_CSV_FILE, IDC_RADIO_TSV_FILE, IDC_RADIO_CSV_FILE);
                    break;
                case txt_file:
                    CheckRadioButton( hWnd, IDC_RADIO_CSV_FILE, IDC_RADIO_TSV_FILE, IDC_RADIO_TSV_FILE);
                    break;
                default:
                    break;
            }
            switch( titraxTimeTypeOption )
            {
                case time_excel:
                    CheckRadioButton( hWnd, IDC_EXCEL_TIME_SERIAL, IDC_SECONDS_SINCE_1970, IDC_EXCEL_TIME_SERIAL);
                    break;
                case time_string:
                    CheckRadioButton( hWnd, IDC_EXCEL_TIME_SERIAL, IDC_SECONDS_SINCE_1970, IDC_LOCALE_TIME_STRING);
                    break;
                case time_1900:
                    CheckRadioButton( hWnd, IDC_EXCEL_TIME_SERIAL, IDC_SECONDS_SINCE_1970, IDC_SECONDS_SINCE_1900);
                    break;
                case time_1904:
                    CheckRadioButton( hWnd, IDC_EXCEL_TIME_SERIAL, IDC_SECONDS_SINCE_1970, IDC_SECONDS_SINCE_1904);
                    break;
                case time_1970:
                    CheckRadioButton( hWnd, IDC_EXCEL_TIME_SERIAL, IDC_SECONDS_SINCE_1970, IDC_SECONDS_SINCE_1970);
                    break;
                default:
                    break;
            }
            if( titraxDumpHistory == 1)
                CheckDlgButton(hWnd,IDC_HISTORY,BST_CHECKED);
            if( titraxDumpSummary ==1 )
                CheckDlgButton(hWnd,IDC_SUMMARY,BST_CHECKED);
            break;
        case WM_COMMAND:
            switch (wParam) {  // something happened inside our dialog

                case IDC_RADIO_CSV_FILE:
                    titraxFileTypeOption = csv_file;
                    break;
                case IDC_RADIO_TSV_FILE:
                    titraxFileTypeOption = txt_file;
                    break;
                case IDC_EXCEL_TIME_SERIAL:
                    titraxTimeTypeOption = time_excel;
                    break;
                case IDC_LOCALE_TIME_STRING:
                    titraxTimeTypeOption = time_string;
                    break;
                case IDC_SECONDS_SINCE_1900:
                    titraxTimeTypeOption = time_1900;
                    break;
                case IDC_SECONDS_SINCE_1904:
                    titraxTimeTypeOption = time_1904;
                    break;
                case IDC_SECONDS_SINCE_1970:
                    titraxTimeTypeOption = time_1970;
                    break;

                case IDC_HISTORY:
                    if( IsDlgButtonChecked(hWnd,IDC_HISTORY) == BST_CHECKED )
                        titraxDumpHistory=TRUE;
                    else
                        titraxDumpHistory=FALSE;
                    break;
                case IDC_SUMMARY:
                    if( IsDlgButtonChecked(hWnd,IDC_SUMMARY) == BST_CHECKED )
                        titraxDumpSummary=TRUE;
                    else
                        titraxDumpSummary=FALSE;
                    break;

                case IDCANCEL:
                    // revert to defaults
                    titraxReadRegistry();
                    EndDialog(hWnd, 1);
                    return TRUE;
                case IDOK:
                    // Save current state
                    titraxWriteRegistry();
                    EndDialog(hWnd, 0);
                    return TRUE;
                case IDHELP: //user pressed dialogue help
                    return DoHelpFile(szPath);
                default:
                    break;
            }
            break;
        case WM_HELP:
            return DoHelpFile(szPath); //user pressed f1 or menu help (from windows not our dialogue)
        default:
            break;
    }
    return FALSE;
}

/*
*****************************************************************************
*     Method:        DoHelpFile
*
*     Description:    Show built in online help
*
*     Parameters(1):    TCHAR* szPath - string path. directory where help file is located
*
*     Returns:        always TRUE for now.

*****************************************************************************
*/
boolean DoHelpFile( TCHAR* szPath )
{
    // find help; where the conduit is
    if (GetModuleFileName((HINSTANCE)hAppInstance, szPath, sizeof(szPath)))
    {
        // Look for the end of the path
        int fileOffset;
        fileOffset = strlen(szPath);
        for (; fileOffset > 0; fileOffset--)
        {
            if (szPath[fileOffset] == '\\')
                break;
        }
        if (fileOffset > 0)
        {
            //szPath[fileOffset + 1] = '\0';
            //StringCbCatEx(szPath,sizeof(szPath)
            //strcat(szPath, IDS_HELP_FILE);
            //TODO
            MessageBox(GetFocus(), "Help isn't implemented yet. Need to figure out .hlp or html help compiler." , "UserPressedHelp", MB_OK | MB_ICONQUESTION );
            //WinHelp(hWnd, szPath, HELP_CONTEXT, IDD_CONDUIT_ACTION + 0x20000 /*base resource*/);
        }
    }
    return TRUE;
}



/*
*****************************************************************************
*
*     Method:        loadTitraxRecords
*
*     Description:    Read the titrax names and data DB's and load the records[]
*                     structure for use by some other routines.
*
*     Parameters(0):  
*
*     Returns:        
*           !0 error
*****************************************************************************
*/
int loadTitraxRecords()
{
    BYTE rHandle;
    long err;
    bool dbOpen = false;
    int i;


    // pretty date stuff
    int tmpTime=0,elapsedMins=0,elapsedHours=0,elapsedDays=0,totalTime=0;

    // open names (which loads recordCount for upcoming looops)
    loadTitraxNames();

    // open the slots database
    if ((err = SyncOpenDB(titraxSlotDBName, 0, rHandle, 
        eDbRead | eDbShowSecret )) != 0) {
        LogAddEntry("SyncOpenDB failed", slWarning, false);
        goto exit;
    }
    dbOpen = true;

    CRawRecordInfo recordInfo;
    recordInfo.m_FileHandle = rHandle;

    for (i = 0; i < recordCount; i++) {

        recordInfo.m_RecIndex = i;
        recordInfo.m_TotalBytes = (unsigned short) sizeof(gBigBuffer);
        recordInfo.m_RecSize = (unsigned short) sizeof(gBigBuffer);
        recordInfo.m_pBytes = (BYTE *) gBigBuffer;
        recordInfo.m_dwReserved = 0;
        
        if ( (err = SyncReadRecordByIndex(recordInfo)) != 0 ) {
            LogAddEntry("SyncReadRecordByIndex failed", slWarning, false);
            goto exit;
        }   

        if( recordInfo.m_RecSize > recordInfo.m_TotalBytes ){
            LogAddEntry("The gBigBuffer was to small to read a record.", slWarning, false);
            goto exit;
        }

        ASSERT( recordInfo.m_RecSize != 0 );

        packedTitraxSlotDataType slot;
        memcpy( (void *) &slot, (void *) recordInfo.m_pBytes, recordInfo.m_RecSize );
        
        (records[i]).index = i;
        (records[i]).elapsed = SyncHHToHostDWord(slot.elapsed);
        (records[i]).startedAt = SyncHHToHostDWord(slot.startedAt);
        (records[i]).active = slot.active;
        (records[i]).logging = slot.logging;
        (records[i]).ignored = slot.ignored;    
 
    }


    if (dbOpen)
        if ((err = SyncCloseDB(rHandle)) != 0)
            LogAddEntry("SyncDBClose failed", slWarning, false);
        else
            dbOpen=false;


    // open the notes database
    if ((err = SyncOpenDB(titraxNoteDBName, 0, rHandle, 
        eDbRead | eDbShowSecret )) != 0) {
        LogAddEntry("SyncOpenDB failed", slWarning, false);
        goto exit;
    }
    dbOpen = true;

    recordInfo.m_FileHandle = rHandle;

    for (i = 0; i < recordCount; i++) {

        recordInfo.m_RecIndex = i;
        recordInfo.m_TotalBytes = (unsigned short) sizeof(gBigBuffer);
        recordInfo.m_RecSize = (unsigned short) sizeof(gBigBuffer);
        recordInfo.m_pBytes = (BYTE *) gBigBuffer;
        recordInfo.m_dwReserved = 0;
        
        if ((err = SyncReadRecordByIndex(recordInfo)) !=0) {
            LogAddEntry("SyncReadRecordByIndex failed", slWarning, false);
            goto exit;
        }   

        if( recordInfo.m_RecSize > recordInfo.m_TotalBytes ){
            LogAddEntry("The gBigBuffer was to small to read a record.", slWarning, false);
            goto exit;
        }

        ASSERT( recordInfo.m_RecSize != 0 );

        memcpy( (void *) &(records[i].note), (void *) recordInfo.m_pBytes, recordInfo.m_RecSize );
 
    }

    if (dbOpen)
        if ((err = SyncCloseDB(rHandle)) != 0)
            LogAddEntry("SyncDBClose failed", slWarning, false);
        else
            dbOpen=false;


    // at this point the array of records should be complete. We can do stuff with it.
    // later think about the GUI having options that get checked here for different
    // types of output

exit:
    if (dbOpen)
        if ((err = SyncCloseDB(rHandle)) != 0)
            LogAddEntry("SyncDBClose failed", slWarning, false);
    return err;

}

/*
*****************************************************************************
*
*     Method:        loadTitraxRecords
*
*     Description:    read the names from titraxNamesDb. The names will 
*                     give us a total record count and their respective indexes
*                     in the list.
*
*     Parameters(0):
*
*     Returns: 
*              !0 - error
*
*****************************************************************************
*/
DWORD loadTitraxNames()
{
    
    BYTE rHandle;
    long err;
    bool dbOpen = false;
    int i;
    WORD thePalmsRecordCount=0;

    recordCount = 0;

    // open the names database
    if ((err = SyncOpenDB(titraxNameDBName, 0, rHandle, 
        eDbRead | eDbShowSecret )) != 0) {
        LogAddEntry("SyncOpenDB failed", slWarning, false);
        goto exit;
    }
    dbOpen = true;

    // get the record count
    if ((err = SyncGetDBRecordCount(rHandle, thePalmsRecordCount)) != 0) {
        LogAddEntry("SyncGetDBRecordCount failed", slWarning, false);
        goto exit;
    }

    CRawRecordInfo recordInfo;
    recordInfo.m_FileHandle = rHandle;

    for (i = 0; i < thePalmsRecordCount; i++) {

        recordInfo.m_RecIndex = i;
        recordInfo.m_TotalBytes = (unsigned short) sizeof(gBigBuffer);
        recordInfo.m_RecSize = (unsigned short) sizeof(gBigBuffer);
        recordInfo.m_pBytes = (BYTE*) gBigBuffer;
        recordInfo.m_dwReserved = 0;
        
        if ((err = SyncReadRecordByIndex(recordInfo)) != 0) {
            LogAddEntry("SyncReadRecordByIndex failed", slWarning, false);
            goto exit;
        }   

        if( recordInfo.m_RecSize > recordInfo.m_TotalBytes ){
            LogAddEntry("The gBigBuffer was to small to read a record.", slWarning, false);
            goto exit;
        }

        ASSERT( recordInfo.m_RecSize != 0 );

        if( recordInfo.m_RecSize > 1 ){ // all records have \0 in them and are at least 1
            memcpy( (void *) &(records[recordCount].name), (void *) recordInfo.m_pBytes, recordInfo.m_RecSize );
            recordCount++;
        }
 
    }

exit:
    if (dbOpen)
        if ((err = SyncCloseDB(rHandle)) != 0)
            LogAddEntry("SyncDBClose failed", slWarning, false);
    return err;

}



/*
*****************************************************************************
*
*     Method:        resString
*
*     Description:    COnvert resourceId to string pointer.
*                     return a pointer to a static buffer that has the string from the resource id in it.
*                     NOTE: STATIC BUFFER. Be quick about it and only do copies. Don't deal in pointers.
*
*     Parameters(1): UINT stringResID - the id number of the string resource
*
*     Returns: 
*              TCHAR* - pointer to the resource string in a static buffer
*
*****************************************************************************
*/
TCHAR* resString( UINT stringResId )
{
    static TCHAR resStringBuffer[MAX_PATH];
    LoadString((HINSTANCE)hLangInstance, stringResId, resStringBuffer, MAX_PATH);
    return resStringBuffer;
}


/*
*****************************************************************************
*
*     Method:        elapsedSecondsToDDDHHMMSS
*
*     Description:    Convert elapsed seconds to  days, hours, minutes and seconds.
*                     NOTE: Only for times less than a year
*
*     Parameters(2): 
*                       DWORD elapsed - seconds since 1/1/1970 00:00:00
*                       TCHAR* whereYouWantIt - TCHAR buffer to hold time. Win32 time string is always
*                       11 characters long.
*
*     Returns: void
*              
*
*****************************************************************************
*/
void elapsedSecondsToDDDHHMMSS( DWORD elapsed, TCHAR* whereYouWantIt )
{
    memset((void *)whereYouWantIt,'\0',11);

    // pretty date stuff
    //int tmpTime=0,elapsedMins=0,elapsedHours=0,elapsedDays=0,elapsedSecs=0;
    tm* tmThing = gmtime( (time_t *) &elapsed);

    StringCchPrintf(whereYouWantIt,11,"%d:%d:%d:%d",
                        tmThing->tm_yday,
                        tmThing->tm_hour,
                        tmThing->tm_min
                        );

}



/*
*****************************************************************************
*
*     Method:        titraxDumpHistory
*
*     Description:    read titraxHistDB and output to a history log file
*
*     Parameters(1): TCHAR* pathName - the directory to place the file in.
*
*     Returns: 
*              !0 - error
*
*****************************************************************************
*/
DWORD titraxDumpHistoryToFile(TCHAR *pathName)
{
    FILE* fp;
    BYTE rHandle;
    long err;
    bool dbOpen = false;
    int i=0;
    WORD thePalmsRecordCount=0;
    TCHAR fileNameBuffer[MAX_PATH];
    packedTitraxHistDataType theRec;
    time_t strt=0,stp=0;


    // open our output file
    //strcpy(fileNameBuffer, pathName );
    //strcat(fileNameBuffer, titraxHistDBName );
    //strcat(fileNameBuffer, IDS_CSV_FILE_EXTENSION );
    StringCchCopy(fileNameBuffer,sizeof(fileNameBuffer),pathName);
    StringCchCat(fileNameBuffer,sizeof(fileNameBuffer),titraxHistDBName);

    switch( titraxFileTypeOption )
    {
        case csv_file:
            StringCchCat(fileNameBuffer,sizeof(fileNameBuffer),resString(IDS_CSV_FILE_EXTENSION));
            break;
        case txt_file:
            StringCchCat(fileNameBuffer,sizeof(fileNameBuffer),resString(IDS_TXT_FILE_EXTENSION));
            break;
    }

    // open the history database
    if ((err = SyncOpenDB(titraxHistDBName, 0, rHandle, eDbRead | eDbShowSecret )) != 0) {
        LogAddEntry("SyncOpenDB titraxHistoryDBName failed", slWarning, false);
        goto exit;
    }
    dbOpen = true;

    // get the record count.
    if ((err = SyncGetDBRecordCount(rHandle, thePalmsRecordCount)) != 0) {
        LogAddEntry("SyncGetDBRecordCount failed", slWarning, false);
        goto exit;
    }


    CRawRecordInfo recordInfo;
    recordInfo.m_FileHandle = rHandle;


    if ((fp = fopen(fileNameBuffer, "w")) == NULL) {
        LogAddFormattedEntry(slWarning, false, "fopen(%s) failed",fileNameBuffer);
        goto exit;
    }

    // header. Magic string. Ick.
    switch( titraxFileTypeOption )
    {
        case csv_file:
            fprintf(fp,"\"name\",\"start\",\"stop\"\n");
            break;
        case txt_file:
            fprintf(fp,"name\tstart\tstop\n");
            break;
    }
    for (i = 0; i < thePalmsRecordCount; i++) {

        recordInfo.m_RecIndex = i;
        recordInfo.m_TotalBytes = (unsigned short) sizeof(gBigBuffer);
        recordInfo.m_RecSize = (unsigned short) sizeof(gBigBuffer);
        recordInfo.m_pBytes = (BYTE *) gBigBuffer;
        recordInfo.m_dwReserved = 0;
        
        if ((err = SyncReadRecordByIndex(recordInfo)) != 0) {
            LogAddEntry("SyncReadRecordByIndex failed", slWarning, false);
            goto exit;
        }   

        if( recordInfo.m_RecSize > recordInfo.m_TotalBytes ){
            LogAddEntry("The gBigBuffer was to small to read a record.", slWarning, false);
            goto exit;
        }

        ASSERT( recordInfo.m_RecSize != 0 );

        if( recordInfo.m_RecSize > 0 ){


            // this way reads the whole record but has byte alignment problems
            // with prc-tools which does 2-byte alignment versus this DLL which 
            // has 4 byte alignment. It's a complex problem with a variety of solutions
            // at different levels. I'll look...

            // Answer: set /Zp2 (2 byte alignment in desktop to match m68k)

            /*
                ms-help://MS.VSCC.2003/MS.MSDNQTR.2003FEB.1033/vccore/html/_core_.2f.Zp.htm

                To set this compiler option in the Visual Studio development environment 

                Open the project's Property Pages dialog box. 
                Click the C/C++ folder. 
                Click the Code Generation property page. 
                Modify the Struct Member Alignment property. 
            */
            memcpy( (void *) &theRec, (void *) recordInfo.m_pBytes, recordInfo.m_RecSize );

            // this is deprecated by the above...
            // try it per sub data type? Meausure from the end of the memory structure
            // This seems to work best. The above can work, but only
            // if the string length of histType.name is divisible by 4 (and 2) to line
            // up with both the intel and m68k byte alignment tendencies. Are these settable in the
            // compiler?
            // name string (it's at the beginning of the structure and all \0's after (usually?)
            //memcpy( (void *) &(theRec.name), (void *) recordInfo.m_pBytes, MAXTEXT+1 );
            // start uint32 (it's the second to last DWORD
            //memcpy( (void *) &(theRec.start), recordInfo.m_pBytes + recordInfo.m_RecSize - 2*sizeof(DWORD) , sizeof(DWORD));
            // stop uint32 (it's the last DWORD
            //memcpy( (void *) &(theRec.stop), recordInfo.m_pBytes + recordInfo.m_RecSize - sizeof(DWORD), sizeof(DWORD) );


            strt = SyncHHToHostDWord(theRec.start);
            stp = SyncHHToHostDWord(theRec.stop);
            //strt = PalmTime2WinTime(SyncHHToHostDWord(theRec.start));
            //stp = PalmTime2WinTime(SyncHHToHostDWord(theRec.stop));
            
            TCHAR startString[MAX_PATH];
            TCHAR stopString[MAX_PATH];
            switch( titraxTimeTypeOption )
            {
                case time_excel:
                    StringCchPrintf(startString,MAX_PATH,"%lf",PalmTime2Excel1900TimeSerial(strt));
                    StringCchPrintf(stopString,MAX_PATH,"%lf",PalmTime2Excel1900TimeSerial(stp));
                    break;
                case time_string:
                    strt = PalmTime2WinTime(strt);
                    stp = PalmTime2WinTime(stp);
                    StringCchPrintf(startString,MAX_PATH,"%0.24s",asctime(gmtime(&strt)));
                    StringCchPrintf(stopString,MAX_PATH,"%0.24s",asctime(gmtime(&stp)));
                    break;
                case time_1900:
                    StringCchPrintf(startString,MAX_PATH,"%lu",PalmTimeTo1900Time(strt));
                    StringCchPrintf(stopString,MAX_PATH,"%lu",PalmTimeTo1900Time(stp));
                    break;
                case time_1904:
                    StringCchPrintf(startString,MAX_PATH,"%lu",strt);
                    StringCchPrintf(stopString,MAX_PATH,"%lu",stp);
                    break;
                case time_1970:
                    StringCchPrintf(startString,MAX_PATH,"%lu",PalmTime2WinTime(strt));
                    StringCchPrintf(stopString,MAX_PATH,"%lu",PalmTime2WinTime(stp));
                    break;
                default:
                    StringCchPrintf(startString,MAX_PATH,"%lu",PalmTime2Excel1900TimeSerial(strt));
                    StringCchPrintf(stopString,MAX_PATH,"%lu",PalmTime2Excel1900TimeSerial(stp));
                    break;
            }
            switch( titraxFileTypeOption )
            {
                case csv_file:
                    fprintf(fp,
                            "\"%s\",\"%s\",\"%s\"\n",
                            theRec.name,
                            startString,
                            stopString
                            );
                    break;
                case txt_file:
                    fprintf(fp,
                            "\"%s\"\t\"%s\"\t\"%s\"\n",
                            theRec.name,
                            startString,
                            stopString
                            );
                    break;
                default:
                    fprintf(fp,
                            "\"%s\",%s,%s\n",
                            theRec.name,
                            startString,
                            stopString
                            );
                    break;

            }
        }
 
    }



exit:
    if (dbOpen)
        if ((err = SyncCloseDB(rHandle)) != 0)
            LogAddEntry("SyncDBClose failed", slWarning, false);
    if(fp != NULL)
            fclose(fp);
    return err;

    return 0;
}


/*
*****************************************************************************
*
*     Method:        titraxDumpSummary
*
*     Description:    dump time values of records[] to file
*
*     Parameters(1): TCHAR* pathName - the directory to place the file in.
*
*     Returns: 
*              !0 - error
*
*****************************************************************************
*/
DWORD titraxDumpSummaryToFile(TCHAR *pathName)
{
    FILE *fp = NULL;
    long err = 0;
    TCHAR fileNameBuffer[MAX_PATH];
    int i=0,totalTime=0;

    // open our output file
    //strcpy(fileNameBuffer, pathName );
    //strcat(fileNameBuffer, titraxNameDBName );
    //strcat(fileNameBuffer, IDS_CSV_FILE_EXTENSION );
    StringCchCopy(fileNameBuffer,sizeof(fileNameBuffer),pathName);
    StringCchCat(fileNameBuffer,sizeof(fileNameBuffer),titraxNameDBName);

    switch( titraxFileTypeOption )
    {
        case csv_file:
            StringCchCat(fileNameBuffer,sizeof(fileNameBuffer),resString(IDS_CSV_FILE_EXTENSION));
            break;
        case txt_file:
            StringCchCat(fileNameBuffer,sizeof(fileNameBuffer),resString(IDS_TXT_FILE_EXTENSION));
            break;
    }

    if ((fp = fopen(fileNameBuffer, "w")) == NULL) {
        LogAddFormattedEntry(slWarning, false, "fopen(%s) failed",
            fileNameBuffer);
        goto exit;
    }

    // header. Magic string. Ick.
    switch( titraxFileTypeOption )
    {
        case csv_file:
            fprintf( fp,"\"Project Name\",\"Time\",\"Notes\"\n" );
            break;
        case txt_file:
            fprintf( fp,"Project Name\tTime\tNotes\n" );
            break;
    }





    for(i=0;i<recordCount;i++){

        TCHAR timeString[MAX_PATH];
        switch( titraxTimeTypeOption )
        {
            case time_excel:
                StringCchPrintf(timeString,MAX_PATH,"%lf",ElapsedTime2Excel1900TimeSerial(records[i].elapsed));
                break;
            case time_string:
                TCHAR tmpString[11];
                elapsedSecondsToDDDHHMMSS( records[i].elapsed, tmpString );
                StringCchPrintf(timeString,MAX_PATH,"%s",tmpString);
                break;
            case time_1900: // elapsed seconds are the same no matter what starting point
            case time_1904:
            case time_1970:
            default:
                StringCchPrintf(timeString,MAX_PATH,"%lu",records[i].elapsed);
                break;
        }
        switch( titraxFileTypeOption )
        {
            case csv_file:
                fprintf(fp,"\"%s\",\"%s\",\"%s\"\n",
                            records[i].name,
                            timeString,
                            records[i].note
                            );
                break;
            case txt_file:
                fprintf(fp,"\"%s\"\t\"%s\"\t\"%s\"\n",
                            records[i].name,
                            timeString,
                            records[i].note
                            );
                break;
        }
    }
exit:
    if (fp != NULL)
        fclose(fp);

    return err;
}


/*
*****************************************************************************
*
*     Function:      PalmTime2WinTime()
*
*     Description:   palm time is the number of seconds since 12a jan 1 1904
*                    windows time is is 12a Jan 1 1970
*
*     Parameters(1):
*                      palmTime - a DWORD  palm time
*
*     Returns:      a time_t windows time
*
*****************************************************************************
*/
time_t PalmTime2WinTime( DWORD palmTime ){
    time_t d;
    d = palmTime - 2082844800; // offset from 1970 to 1904 in seconds
    return  d;
}


/*
*****************************************************************************
*
*     Function:      WinTime2PalmTime()
*
*     Description:   palm time is the number of seconds since 12a jan 1 1904
*                    windows time is is 12a Jan 1 1970
*
*     Parameters(1):
*                      winTime - a time in windows format
*
*     Returns:      a DWORD representing palm time
*
*****************************************************************************
*/
DWORD WinTime2PalmTime( time_t winTime ){
    DWORD d;
    d = winTime + 2082844800; // offset from 1970 to 1904 in seconds
    return  d;
}


/*
*****************************************************************************
*
*     Function:      ElapsedTime2Excel1900TimeSerial()
*
*     Description:      Make a time serial from elapsed time. No need for
*                       date conversion.
*
*
*     Parameters(1):
*                      elapsedTime - an elapsed time format
*
*     Returns:      a double representing excel time
*
*****************************************************************************
*/
double ElapsedTime2Excel1900TimeSerial( time_t elapsedTime )
{
    double excelTime;
    UINT32 days;
    UINT32 remainderSeconds;

    // whole number remainder seconds
    remainderSeconds = elapsedTime % 86400;

    // fraction: remainders seconds part of a day
    excelTime = ( (double)remainderSeconds / (double) 86400 );

    // whole number days
    excelTime += (double)( elapsedTime / 86400  ); // palmTime / 60 / 60 / 24
    
    return excelTime;
}


/*
*****************************************************************************
*
*     Function:      PalmTime2Excel1900TimeSerial()
*
*     Description:      Excel time serials are floating point values. The whole part is days.
*                       and the fractional part is percentages of a day. All measured from
*                       1900 (or 1904 using options).  Since the fractional part is percentage of a day (it's a ratio)
*                       and you can use any unit you want as long as the percentage calculations
*                       use the same unit.
*                          
*                       Palm Time is the number of seconds since 12a jan 1 1904.
*
*                       number of days (including one leap day) differences is 1,462 * 24 * 60 * 60 = 126316800 
*
*
*     Parameters(1):
*                      palmTime - a time in palm format
*
*     Returns:      a double representing excel time
*
*****************************************************************************
*/
double PalmTime2Excel1900TimeSerial( time_t palmTime )
{
    double excelTime;
    UINT32 days;
    UINT32 remainderSeconds;

    // change from 1904 to 1900 (get's bigger to cover previously uncounted 4 years)
    palmTime+=126316800;

    // whole number remainder seconds
    remainderSeconds = palmTime % 86400;

    // fraction: remainders seconds part of a day
    excelTime = ( (double)remainderSeconds / (double) 86400 );

    // whole number days
    excelTime += (double)( palmTime / 86400  ); // palmTime / 60 / 60 / 24
    
    return excelTime;
}

/*
*****************************************************************************
*
*     Function:      PalmTimeTo1900Time
*
*     Description:      Convert a palm 1904 time to a 1900 time.
*                         1900 was a leap year. So one extra day.
*                         4 years = (4 * 365 * 24 * 60 * 60) + (24 * 60 * 60) = 126230400
*     Parameters(1):
*                      time_t palmTime - a time in palm format
*
*     Returns:      a DWORD representing 1900 time
*
*****************************************************************************
*/
DWORD PalmTimeTo1900Time( DWORD palmTime )
{
    // time gets bigger to account for previously uncounted years
    palmTime+=126230400;
    return palmTime;
}

/*
*****************************************************************************
*
*     Function:      WinTime2Excel1900TimeSerial
*
*     Description:      Excel time serials are floating point values. The whole part is days.
*                       and the fractional part is percentages of a day. All measured from
*                       1900 (or 1904 using options).
*                       Just a convenience wrapper for other calls. no math.
*
*     Parameters(1):
*                      time_t winTime - a time in win format
*
*     Returns:      a double representing excel time
*
*****************************************************************************
*/
double WinTime2Excel1900TimeSerial( time_t winTime ){
    return PalmTime2Excel1900TimeSerial(WinTime2PalmTime(winTime));
}


/*
*****************************************************************************
*
*     Function:      titraxWriteRegistry
*
*     Description:   Store user settings to registry
*
*     Parameters(0):
*
*     Returns:      !0 is error code
*
*****************************************************************************
*/
DWORD titraxWriteRegistry(void)
{
    DWORD err=0;
    DWORD val=0;
    HKEY hkey;

    // open registry
    err = RegOpenKeyEx( HKEY_CURRENT_USER, resString(IDS_REGISTRY_PATH),NULL, KEY_WRITE,&hkey);
    if( err != ERROR_SUCCESS )
        return 1;
    
    // write history setting
    val = titraxDumpHistory;
    err = RegSetValueEx(hkey,"TitraxOutputHistory",NULL,REG_DWORD,(BYTE *)&val,sizeof(val));
    if( err != ERROR_SUCCESS )
        return 1;

    // write summary setting
    val = titraxDumpSummary;
    err = RegSetValueEx(hkey,"TitraxOutputSummary",NULL,REG_DWORD,(BYTE *)&val,sizeof(val));
    if( err != ERROR_SUCCESS )
        return 1;

    // write file type preference
    val = titraxFileTypeOption;
    err = RegSetValueEx(hkey,"TitraxFileFormat",NULL,REG_DWORD,(BYTE *)&val,sizeof(val));
    if( err != ERROR_SUCCESS )
        return 1;

    // write time type preference
    val = titraxTimeTypeOption;
    err = RegSetValueEx(hkey,"TitraxTimeFormat",NULL,REG_DWORD,(BYTE *)&val,sizeof(val));
    if( err != ERROR_SUCCESS )
        return 1;
    
    // close the registry
    err = RegCloseKey( hkey );
    if( err != ERROR_SUCCESS )
        return 1;

    return err;
}

/*
*****************************************************************************
*
*     Function:      titraxReadRegistry
*
*     Description:   load user settings from the registry into global variables
*
*     Parameters(0):
*
*     Returns:      !0 is error code
*
*****************************************************************************
*/
DWORD titraxReadRegistry(void)
{
    DWORD err=0;
    DWORD val=0;
    DWORD readSize;
    HKEY hkey;
    err = RegOpenKey( HKEY_CURRENT_USER,  resString(IDS_REGISTRY_PATH),&hkey);

    if( err != ERROR_SUCCESS )
        return 1;

    // output summary?
    err = RegQueryValueEx( hkey, "TitraxOutputSummary", NULL, NULL,
               (LPBYTE) &val, &readSize);
    if( err != ERROR_SUCCESS || ( readSize > sizeof(val) ) )
        return 1;
    titraxDumpSummary = val;

    // output history?
    err = RegQueryValueEx( hkey, "TitraxOutputHistory", NULL, NULL,
               (LPBYTE) &val, &readSize);
    if( err != ERROR_SUCCESS || ( readSize > sizeof(val) ) )
        return 1;
    titraxDumpHistory = val;

    // file format?
    err = RegQueryValueEx( hkey, "TitraxFileFormat", NULL, NULL,
               (LPBYTE) &val, &readSize);
    if( err != ERROR_SUCCESS || ( readSize > sizeof(val) ) )
        return 1;
    titraxFileTypeOption = (titraxFileType) val;

    // time format?
    err = RegQueryValueEx( hkey, "TitraxTimeFormat", NULL, NULL,
               (LPBYTE) &val, &readSize);
    if( err != ERROR_SUCCESS || ( readSize > sizeof(val) ) )
        return 1;
    titraxTimeTypeOption = (titraxTimeType) val;

    err = RegCloseKey(hkey);
    if( err != ERROR_SUCCESS )
        return 1;

    return err;
}
